@@ -6,7 +6,7 @@
  * This file is part of Libgcrypt.
  *
  * Libgcrypt is free software; you can redistribute it and/or modify
- * it under the terms of the GNU Lesser general Public License as
+ * it under the terms of the GNU Lesser General Public License as
  * published by the Free Software Foundation; either version 2.1 of
  * the License, or (at your option) any later version.
  *
@@ -48,6 +48,7 @@ static gcry_pk_spec_t * const pubkey_list[] =
 #if USE_ELGAMAL
     &_gcry_pubkey_spec_elg,
 #endif
+    &_gcry_pubkey_spec_kem,
     NULL
   };
 
@@ -327,8 +328,17 @@ _gcry_pk_encrypt (gcry_sexp_t *r_ciph, gcry_sexp_t s_data, gcry_sexp_t s_pkey)
   if (spec->flags.disabled)
     rc = GPG_ERR_PUBKEY_ALGO;
   else if (!spec->flags.fips && fips_mode ())
-    rc = GPG_ERR_PUBKEY_ALGO;
-  else if (spec->encrypt)
+    {
+      if (fips_check_rejection (GCRY_FIPS_FLAG_REJECT_PK))
+        {
+          rc = GPG_ERR_PUBKEY_ALGO;
+          goto leave;
+        }
+      else
+        fips_service_indicator_mark_non_compliant ();
+    }
+
+  if (spec->encrypt)
     rc = spec->encrypt (r_ciph, s_data, keyparms);
   else
     rc = GPG_ERR_NOT_IMPLEMENTED;
@@ -440,8 +450,17 @@ _gcry_pk_sign (gcry_sexp_t *r_sig, gcry_sexp_t s_hash, gcry_sexp_t s_skey)
   if (spec->flags.disabled)
     rc = GPG_ERR_PUBKEY_ALGO;
   else if (!spec->flags.fips && fips_mode ())
-    rc = GPG_ERR_PUBKEY_ALGO;
-  else if (spec->sign)
+    {
+      if (fips_check_rejection (GCRY_FIPS_FLAG_REJECT_PK))
+        {
+          rc = GPG_ERR_PUBKEY_ALGO;
+          goto leave;
+        }
+      else
+        fips_service_indicator_mark_non_compliant ();
+    }
+
+  if (spec->sign)
     rc = spec->sign (r_sig, s_hash, keyparms);
   else
     rc = GPG_ERR_NOT_IMPLEMENTED;
@@ -452,6 +471,157 @@ _gcry_pk_sign (gcry_sexp_t *r_sig, gcry_sexp_t s_hash, gcry_sexp_t s_skey)
 }
 
 
+#define MAX_CONTEXTS 2 /* Currently, input data and random_override */
+
+static gcry_err_code_t
+prepare_datasexp_to_be_signed (const char *tmpl, gcry_md_hd_t hd,
+                               gcry_ctx_t ctx, gcry_sexp_t *s_data_p)
+{
+  const char *s;
+  const char *digest_name = NULL;
+  const unsigned char *digest;
+  int digest_size;
+  int algo;
+  gcry_err_code_t rc;
+
+  if (hd == NULL)
+    {
+      const unsigned char *data[MAX_CONTEXTS];
+      int data_size[MAX_CONTEXTS];
+      int i = 0;
+      void *argv[MAX_CONTEXTS*2];
+
+      while (1)
+        {
+          size_t len;
+
+          rc = _gcry_pk_get_single_data (&ctx, &data[i/2], &len);
+          if (rc)
+            return rc;
+
+          data_size[i/2] = (int)len;
+
+          argv[i] = (void *)&data_size[i/2];
+          argv[i+1] = (void *)&data[i/2];
+
+          i += 2;
+
+          if (!ctx)
+            break;
+
+          if (i >= MAX_CONTEXTS*2)
+            return GPG_ERR_EINVAL;
+        }
+
+      rc = _gcry_sexp_build_array (s_data_p, NULL, tmpl, argv);
+      return rc;
+    }
+
+  /* Check if it has fixed hash name or %s */
+  s = strstr (tmpl, "(hash ");
+  if (s == NULL)
+    {
+      _gcry_md_close (hd);
+      return GPG_ERR_DIGEST_ALGO;
+    }
+
+  s += 6;
+  if (!strncmp (s, "%s", 2))
+    {
+      algo = _gcry_md_get_algo (hd);
+
+      if (fips_mode () && algo == GCRY_MD_SHA1)
+        {
+          if (fips_check_rejection (GCRY_FIPS_FLAG_REJECT_PK))
+            {
+              _gcry_md_close (hd);
+              return GPG_ERR_DIGEST_ALGO;
+            }
+          else
+            fips_service_indicator_mark_non_compliant ();
+        }
+
+      digest_name = _gcry_md_algo_name (algo);
+      digest_size = (int)_gcry_md_get_algo_dlen (algo);
+      digest = _gcry_md_read (hd, 0);
+    }
+  else
+    {
+      const char *p;
+      char *digest_name_supplied;
+
+      for (p = s; *p && *p != ' '; p++)
+	;
+
+      digest_name_supplied = xtrymalloc (p - s + 1);
+      if (!digest_name_supplied)
+        {
+          rc = gpg_err_code_from_syserror ();
+          _gcry_md_close (hd);
+          return rc;
+        }
+      memcpy (digest_name_supplied, s, p - s);
+      digest_name_supplied[p - s] = 0;
+
+      algo = _gcry_md_map_name (digest_name_supplied);
+      xfree (digest_name_supplied);
+      if (algo == 0)
+	{
+	  _gcry_md_close (hd);
+	  return GPG_ERR_DIGEST_ALGO;
+	}
+      else if (fips_mode () && algo == GCRY_MD_SHA1)
+        {
+          if (fips_check_rejection (GCRY_FIPS_FLAG_REJECT_PK))
+            {
+              _gcry_md_close (hd);
+              return GPG_ERR_DIGEST_ALGO;
+            }
+          else
+            fips_service_indicator_mark_non_compliant ();
+        }
+
+      digest_size = (int)_gcry_md_get_algo_dlen (algo);
+      digest = _gcry_md_read (hd, algo);
+    }
+
+  if (!digest)
+    {
+      _gcry_md_close (hd);
+      return GPG_ERR_NOT_IMPLEMENTED;
+    }
+
+  if (!ctx)
+    {
+      if (!digest_name)
+	rc = _gcry_sexp_build (s_data_p, NULL, tmpl,
+			       digest_size, digest);
+      else
+	rc = _gcry_sexp_build (s_data_p, NULL, tmpl, digest_name,
+			       digest_size, digest);
+    }
+  else
+    {
+      const unsigned char *p;
+      size_t len;
+
+      rc = _gcry_pk_get_single_data (&ctx, &p, &len);
+      if (rc)
+        return rc;
+
+      if (!digest_name)
+	rc = _gcry_sexp_build (s_data_p, NULL, tmpl,
+                               digest_size, digest, (int) len, p);
+      else
+	rc = _gcry_sexp_build (s_data_p, NULL, tmpl, digest_name,
+                               digest_size, digest, (int) len, p);
+    }
+
+  _gcry_md_close (hd);
+  return rc;
+}
+
+
 gcry_err_code_t
 _gcry_pk_sign_md (gcry_sexp_t *r_sig, const char *tmpl, gcry_md_hd_t hd_orig,
                   gcry_sexp_t s_skey, gcry_ctx_t ctx)
@@ -459,117 +629,22 @@ _gcry_pk_sign_md (gcry_sexp_t *r_sig, const char *tmpl, gcry_md_hd_t hd_orig,
   gcry_err_code_t rc;
   gcry_pk_spec_t *spec;
   gcry_sexp_t keyparms = NULL;
-  gcry_sexp_t s_hash = NULL;
-  int algo;
-  const unsigned char *digest;
-  int digest_size;
+  gcry_sexp_t s_data = NULL;
   gcry_error_t err;
   gcry_md_hd_t hd;
-  const char *s;
-  char *hash_name;
 
   *r_sig = NULL;
 
-  /* Check if it has fixed hash name or %s */
-  s = strstr (tmpl, "(hash ");
-  if (s == NULL)
-    return GPG_ERR_DIGEST_ALGO;
-
-  s += 6;
-  if (!strncmp (s, "%s", 2))
-    hash_name = NULL;
+  if (!hd_orig)
+    hd = NULL;
   else
     {
-      const char *p;
-
-      for (p = s; *p && *p != ' '; p++)
-	;
-
-      hash_name = xtrymalloc (p - s + 1);
-      if (!hash_name)
-	return gpg_error_from_syserror ();
-      memcpy (hash_name, s, p - s);
-      hash_name[p - s] = 0;
+      err = _gcry_md_copy (&hd, hd_orig);
+      if (err)
+        return gpg_err_code (err);
     }
 
-  err = _gcry_md_copy (&hd, hd_orig);
-  if (err)
-    {
-      xfree (hash_name);
-      return gpg_err_code (err);
-    }
-
-  if (hash_name)
-    {
-      algo = _gcry_md_map_name (hash_name);
-      digest_size = (int) _gcry_md_get_algo_dlen (algo);
-
-      if (algo == 0 || digest_size == 0
-          || (fips_mode () && algo == GCRY_MD_SHA1))
-	{
-	  xfree (hash_name);
-	  _gcry_md_close (hd);
-	  return GPG_ERR_DIGEST_ALGO;
-	}
-
-      digest = _gcry_md_read (hd, algo);
-    }
-  else
-    {
-      algo = _gcry_md_get_algo (hd);
-      digest_size = (int) _gcry_md_get_algo_dlen (algo);
-
-      if (digest_size == 0 || (fips_mode () && algo == GCRY_MD_SHA1))
-        {
-          _gcry_md_close (hd);
-          return GPG_ERR_DIGEST_ALGO;
-        }
-
-      digest = _gcry_md_read (hd, 0);
-    }
-
-  if (!digest)
-    {
-      xfree (hash_name);
-      _gcry_md_close (hd);
-      return GPG_ERR_NOT_IMPLEMENTED;
-    }
-
-  if (!ctx)
-    {
-      if (hash_name)
-	rc = _gcry_sexp_build (&s_hash, NULL, tmpl,
-			       digest_size, digest);
-      else
-	rc = _gcry_sexp_build (&s_hash, NULL, tmpl,
-			       _gcry_md_algo_name (algo),
-			       digest_size, digest);
-    }
-  else
-    {
-      const unsigned char *p;
-      size_t len;
-
-      rc = _gcry_pk_get_random_override (ctx, &p, &len);
-      if (rc)
-        {
-          _gcry_md_close (hd);
-          return rc;
-        }
-
-      if (hash_name)
-	rc = _gcry_sexp_build (&s_hash, NULL, tmpl,
-			       digest_size, digest,
-			       (int) len, p);
-      else
-	rc = _gcry_sexp_build (&s_hash, NULL, tmpl,
-			       _gcry_md_algo_name (algo),
-			       digest_size, digest,
-			       (int) len, p);
-    }
-
-  xfree (hash_name);
-  _gcry_md_close (hd);
+  rc = prepare_datasexp_to_be_signed (tmpl, hd, ctx, &s_data);
   if (rc)
     return rc;
 
@@ -580,14 +655,23 @@ _gcry_pk_sign_md (gcry_sexp_t *r_sig, const char *tmpl, gcry_md_hd_t hd_orig,
   if (spec->flags.disabled)
     rc = GPG_ERR_PUBKEY_ALGO;
   else if (!spec->flags.fips && fips_mode ())
-    rc = GPG_ERR_PUBKEY_ALGO;
-  else if (spec->sign)
-    rc = spec->sign (r_sig, s_hash, keyparms);
+    {
+      if (fips_check_rejection (GCRY_FIPS_FLAG_REJECT_PK))
+        {
+          rc = GPG_ERR_PUBKEY_ALGO;
+          goto leave;
+        }
+      else
+        fips_service_indicator_mark_non_compliant ();
+    }
+
+  if (spec->sign)
+    rc = spec->sign (r_sig, s_data, keyparms);
   else
     rc = GPG_ERR_NOT_IMPLEMENTED;
 
  leave:
-  sexp_release (s_hash);
+  sexp_release (s_data);
   sexp_release (keyparms);
   return rc;
 }
@@ -614,8 +698,17 @@ _gcry_pk_verify (gcry_sexp_t s_sig, gcry_sexp_t s_hash, gcry_sexp_t s_pkey)
   if (spec->flags.disabled)
     rc = GPG_ERR_PUBKEY_ALGO;
   else if (!spec->flags.fips && fips_mode ())
-    rc = GPG_ERR_PUBKEY_ALGO;
-  else if (spec->verify)
+    {
+      if (fips_check_rejection (GCRY_FIPS_FLAG_REJECT_PK))
+        {
+          rc = GPG_ERR_PUBKEY_ALGO;
+          goto leave;
+        }
+      else
+        fips_service_indicator_mark_non_compliant ();
+    }
+
+  if (spec->verify)
     rc = spec->verify (s_sig, s_hash, keyparms);
   else
     rc = GPG_ERR_NOT_IMPLEMENTED;
@@ -633,115 +726,20 @@ _gcry_pk_verify_md (gcry_sexp_t s_sig, const char *tmpl, gcry_md_hd_t hd_orig,
   gcry_err_code_t rc;
   gcry_pk_spec_t *spec;
   gcry_sexp_t keyparms = NULL;
-  gcry_sexp_t s_hash = NULL;
-  int algo;
-  const unsigned char *digest;
-  int digest_size;
+  gcry_sexp_t s_data = NULL;
   gcry_error_t err;
   gcry_md_hd_t hd;
-  const char *s;
-  char *hash_name;
 
-  /* Check if it has fixed hash name or %s */
-  s = strstr (tmpl, "(hash ");
-  if (s == NULL)
-    return GPG_ERR_DIGEST_ALGO;
-
-  s += 6;
-  if (!strncmp (s, "%s", 2))
-    hash_name = NULL;
+  if (!hd_orig)
+    hd = NULL;
   else
     {
-      const char *p;
-
-      for (p = s; *p && *p != ' '; p++)
-        ;
-
-      hash_name = xtrymalloc (p - s + 1);
-      if (!hash_name)
-        return gpg_error_from_syserror ();
-      memcpy (hash_name, s, p - s);
-      hash_name[p - s] = 0;
+      err = _gcry_md_copy (&hd, hd_orig);
+      if (err)
+        return gpg_err_code (err);
     }
 
-  err = _gcry_md_copy (&hd, hd_orig);
-  if (err)
-    {
-      xfree (hash_name);
-      return gpg_err_code (err);
-    }
-
-  if (hash_name)
-    {
-      algo = _gcry_md_map_name (hash_name);
-      digest_size = (int) _gcry_md_get_algo_dlen (algo);
-
-      if (algo == 0 || digest_size == 0
-          || (fips_mode () && algo == GCRY_MD_SHA1))
-        {
-          xfree (hash_name);
-          _gcry_md_close (hd);
-          return GPG_ERR_DIGEST_ALGO;
-        }
-
-      digest = _gcry_md_read (hd, algo);
-    }
-  else
-    {
-      algo = _gcry_md_get_algo (hd);
-      digest_size = (int) _gcry_md_get_algo_dlen (algo);
-
-      if (digest_size == 0 || (fips_mode () && algo == GCRY_MD_SHA1))
-        {
-          _gcry_md_close (hd);
-          return GPG_ERR_DIGEST_ALGO;
-        }
-
-      digest = _gcry_md_read (hd, 0);
-    }
-
-  if (!digest)
-    {
-      xfree (hash_name);
-      _gcry_md_close (hd);
-      return GPG_ERR_DIGEST_ALGO;
-    }
-
-  if (!ctx)
-    {
-      if (hash_name)
-        rc = _gcry_sexp_build (&s_hash, NULL, tmpl,
-                               digest_size, digest);
-      else
-        rc = _gcry_sexp_build (&s_hash, NULL, tmpl,
-                               _gcry_md_algo_name (algo),
-                               digest_size, digest);
-    }
-  else
-    {
-      const unsigned char *p;
-      size_t len;
-
-      rc = _gcry_pk_get_random_override (ctx, &p, &len);
-      if (rc)
-        {
-          _gcry_md_close (hd);
-          return rc;
-        }
-
-      if (hash_name)
-        rc = _gcry_sexp_build (&s_hash, NULL, tmpl,
-                               digest_size, digest,
-                               (int) len, p);
-      else
-        rc = _gcry_sexp_build (&s_hash, NULL, tmpl,
-                               _gcry_md_algo_name (algo),
-                               digest_size, digest,
-                               (int) len, p);
-    }
-
-  xfree (hash_name);
-  _gcry_md_close (hd);
+  rc = prepare_datasexp_to_be_signed (tmpl, hd, ctx, &s_data);
   if (rc)
     return rc;
 
@@ -752,14 +750,23 @@ _gcry_pk_verify_md (gcry_sexp_t s_sig, const char *tmpl, gcry_md_hd_t hd_orig,
   if (spec->flags.disabled)
     rc = GPG_ERR_PUBKEY_ALGO;
   else if (!spec->flags.fips && fips_mode ())
-    rc = GPG_ERR_PUBKEY_ALGO;
-  else if (spec->verify)
-    rc = spec->verify (s_sig, s_hash, keyparms);
+    {
+      if (fips_check_rejection (GCRY_FIPS_FLAG_REJECT_PK))
+        {
+          rc = GPG_ERR_PUBKEY_ALGO;
+          goto leave;
+        }
+      else
+        fips_service_indicator_mark_non_compliant ();
+    }
+
+  if (spec->verify)
+    rc = spec->verify (s_sig, s_data, keyparms);
   else
     rc = GPG_ERR_NOT_IMPLEMENTED;
 
  leave:
-  sexp_release (s_hash);
+  sexp_release (s_data);
   sexp_release (keyparms);
   return rc;
 }
@@ -788,8 +795,17 @@ _gcry_pk_testkey (gcry_sexp_t s_key)
   if (spec->flags.disabled)
     rc = GPG_ERR_PUBKEY_ALGO;
   else if (!spec->flags.fips && fips_mode ())
-    rc = GPG_ERR_PUBKEY_ALGO;
-  else if (spec->check_secret_key)
+    {
+      if (fips_check_rejection (GCRY_FIPS_FLAG_REJECT_PK))
+        {
+          rc = GPG_ERR_PUBKEY_ALGO;
+          goto leave;
+        }
+      else
+        fips_service_indicator_mark_non_compliant ();
+    }
+
+  if (spec->check_secret_key)
     rc = spec->check_secret_key (keyparms);
   else
     rc = GPG_ERR_NOT_IMPLEMENTED;
@@ -800,38 +816,34 @@ _gcry_pk_testkey (gcry_sexp_t s_key)
 }
 
 
-/*
-  Create a public key pair and return it in r_key.
-  How the key is created depends on s_parms:
-  (genkey
-   (algo
-     (parameter_name_1 ....)
-      ....
-     (parameter_name_n ....)
-  ))
-  The key is returned in a format depending on the
-  algorithm. Both, private and secret keys are returned
-  and optionally some additional informatin.
-  For elgamal we return this structure:
-  (key-data
-   (public-key
-     (elg
- 	(p <mpi>)
- 	(g <mpi>)
- 	(y <mpi>)
-     )
-   )
-   (private-key
-     (elg
- 	(p <mpi>)
- 	(g <mpi>)
- 	(y <mpi>)
- 	(x <mpi>)
-     )
-   )
-   (misc-key-info
-      (pm1-factors n1 n2 ... nn)
-   ))
+
+/* Create a public key pair and return it as R_KEY.
+ * How the key is created depends on s_parms:
+ *
+ *   (genkey
+ *     (algo
+ *       (parameter_name_1 ....)
+ *       ....
+ *       (parameter_name_n ....)))
+ *
+ * The key is returned in a format depending on the algorithm. Both,
+ * private and secret keys are returned and optionally some additional
+ * information.  For example for Elgamal this structure is returned:
+ *
+ *   (key-data
+ *     (public-key
+ *       (elg
+ *         (p <mpi>)
+ *         (g <mpi>)
+ *         (y <mpi>)))
+ *     (private-key
+ *       (elg
+ *         (p <mpi>)
+ *         (g <mpi>)
+ *         (y <mpi>)
+ *         (x <mpi>)))
+ *     (misc-key-info
+ *        (pm1-factors n1 n2 ... nn)))
  */
 gcry_err_code_t
 _gcry_pk_genkey (gcry_sexp_t *r_key, gcry_sexp_t s_parms)
@@ -871,11 +883,21 @@ _gcry_pk_genkey (gcry_sexp_t *r_key, gcry_sexp_t s_parms)
   spec = spec_from_name (name);
   xfree (name);
   name = NULL;
-  if (!spec || spec->flags.disabled || (!spec->flags.fips && fips_mode ()))
+  if (!spec || spec->flags.disabled)
     {
       rc = GPG_ERR_PUBKEY_ALGO; /* Unknown algorithm.  */
       goto leave;
     }
+  else if (!spec->flags.fips && fips_mode ())
+    {
+      if (fips_check_rejection (GCRY_FIPS_FLAG_REJECT_PK))
+        {
+          rc = GPG_ERR_PUBKEY_ALGO;
+          goto leave;
+        }
+      else
+        fips_service_indicator_mark_non_compliant ();
+    }
 
   if (spec->generate)
     rc = spec->generate (list, r_key);
@@ -911,12 +933,22 @@ _gcry_pk_get_nbits (gcry_sexp_t key)
 
   if (spec_from_sexp (key, 0, &spec, &parms))
     return 0; /* Error - 0 is a suitable indication for that.  */
-  if (spec->flags.disabled)
-    return 0;
-  if (!spec->flags.fips && fips_mode ())
-    return 0;
 
-  nbits = spec->get_nbits (parms);
+  if (spec->flags.disabled)
+    nbits = 0;                  /* Error */
+  else if (!spec->flags.fips && fips_mode ())
+    {
+      if (fips_check_rejection (GCRY_FIPS_FLAG_REJECT_PK))
+        nbits = 0;              /* Error */
+      else
+        {
+          fips_service_indicator_mark_non_compliant ();
+          nbits = spec->get_nbits (parms);
+        }
+    }
+  else
+    nbits = spec->get_nbits (parms);
+
   sexp_release (parms);
   return nbits;
 }
@@ -1049,10 +1081,18 @@ _gcry_pk_get_curve (gcry_sexp_t key, int iterator, unsigned int *r_nbits)
     }
 
   if (spec->flags.disabled)
-    return NULL;
-  if (!spec->flags.fips && fips_mode ())
-    return NULL;
-  if (spec->get_curve)
+    result = NULL;
+  else if (!spec->flags.fips && fips_mode ())
+    {
+      if (fips_check_rejection (GCRY_FIPS_FLAG_REJECT_PK))
+        result = NULL;
+      else
+        {
+          fips_service_indicator_mark_non_compliant ();
+          result = spec->get_curve (keyparms, iterator, r_nbits);
+        }
+    }
+  else if (spec->get_curve)
     result = spec->get_curve (keyparms, iterator, r_nbits);
 
   sexp_release (keyparms);
@@ -1273,47 +1313,50 @@ _gcry_pk_selftest (int algo, int extended, selftest_report_func_t report)
 }
 
 
-struct pk_random_override {
+struct pk_single_data {
   size_t len;
   unsigned char area[1];  /* In future, we may use flexible array member.  */
 };
 
 gpg_err_code_t
-_gcry_pk_random_override_new (gcry_ctx_t *r_ctx,
-                              const unsigned char *p, size_t len)
+_gcry_pk_single_data_push (gcry_ctx_t *r_ctx,
+                          const unsigned char *p, size_t len)
 {
   gcry_ctx_t ctx;
-  struct pk_random_override *pro;
+  struct pk_single_data *psd;
+  int data_type = CONTEXT_TYPE_SINGLE_DATA;
 
-  *r_ctx = NULL;
   if (!p)
     return GPG_ERR_EINVAL;
 
-  ctx = _gcry_ctx_alloc (CONTEXT_TYPE_RANDOM_OVERRIDE,
-                         offsetof (struct pk_random_override, area) + len,
-                         NULL);
+  ctx = _gcry_ctx_alloc (data_type,
+			 offsetof (struct pk_single_data, area) + len,
+                         NULL, *r_ctx);
   if (!ctx)
     return gpg_err_code_from_syserror ();
-  pro = _gcry_ctx_get_pointer (ctx, CONTEXT_TYPE_RANDOM_OVERRIDE);
-  pro->len = len;
-  memcpy (pro->area, p, len);
+  psd = _gcry_ctx_get_pointer (ctx, data_type);
+  psd->len = len;
+  memcpy (psd->area, p, len);
 
   *r_ctx = ctx;
   return 0;
 }
 
 gpg_err_code_t
-_gcry_pk_get_random_override (gcry_ctx_t ctx,
-                              const unsigned char **r_p, size_t *r_len)
+_gcry_pk_get_single_data (gcry_ctx_t *r_ctx,
+                          const unsigned char **r_p, size_t *r_len)
 {
-  struct pk_random_override *pro;
+  struct pk_single_data *psd;
+  int data_type = CONTEXT_TYPE_SINGLE_DATA;
+  gcry_ctx_t ctx = *r_ctx;
 
-  pro = _gcry_ctx_find_pointer (ctx, CONTEXT_TYPE_RANDOM_OVERRIDE);
-  if (!pro)
+  psd = _gcry_ctx_find_pointer (ctx, data_type);
+  if (!psd)
     return GPG_ERR_EINVAL;
 
-  *r_p = pro->area;
-  *r_len = pro->len;
+  *r_p = psd->area;
+  *r_len = psd->len;
+  *r_ctx = _gcry_ctx_get_pointer (ctx, 0);
 
   return 0;
 }
