@@ -39,9 +39,8 @@
 #include "rand-internal.h"
 #include "cipher.h"         /* For _gcry_sha1_hash_buffer().  */
 
-/* The name of a file used to globally configure the RNG.
- * Do not use this macro directly; use get_random_conf_file.  */
-#define RANDOM_CONF_FILE "random.conf"
+/* The name of a file used to globally configure the RNG. */
+#define RANDOM_CONF_FILE "/etc/gcrypt/random.conf"
 
 
 /* If not NULL a progress function called from certain places and the
@@ -89,26 +88,6 @@ _gcry_random_progress (const char *what, int printchar, int current, int total)
 }
 
 
-static const char *
-get_random_conf_file (void)
-{
-#ifdef HAVE_W32_SYSTEM
-  static char *fname;
-
-  if (!fname)
-    {
-      const char *sysconfdir = _gcry_get_sysconfdir();
-
-      fname = xmalloc (strlen (sysconfdir) + strlen (RANDOM_CONF_FILE) + 1);
-      strcpy (fname, sysconfdir);
-      strcat (fname, RANDOM_CONF_FILE);
-    }
-  return fname;
-#else
-  return "/etc/gcrypt/" RANDOM_CONF_FILE;
-#endif
-}
-
 /* Read a file with configure options.  The file is a simple text file
  * where empty lines and lines with the first non white-space
  * character being '#' are ignored.  Supported configure options are:
@@ -123,7 +102,7 @@ get_random_conf_file (void)
 unsigned int
 _gcry_random_read_conf (void)
 {
-  const char *fname = get_random_conf_file ();
+  const char *fname = RANDOM_CONF_FILE;
   FILE *fp;
   char buffer[256];
   char *p, *pend;
@@ -328,28 +307,6 @@ _gcry_enable_quick_random_gen (void)
 }
 
 
-void
-_gcry_set_random_daemon_socket (const char *socketname)
-{
-  if (fips_mode ())
-    ;  /* Not used.  */
-  else
-    _gcry_rngcsprng_set_daemon_socket (socketname);
-}
-
-/* With ONOFF set to 1, enable the use of the daemon.  With ONOFF set
-   to 0, disable the use of the daemon.  With ONOF set to -1, return
-   whether the daemon has been enabled. */
-int
-_gcry_use_random_daemon (int onoff)
-{
-  if (fips_mode ())
-    return 0; /* Never enabled in fips mode.  */
-  else
-    return _gcry_rngcsprng_use_daemon (onoff);
-}
-
-
 /* This function returns true if no real RNG is available or the
    quality of the RNG has been degraded for test purposes.  */
 int
