@@ -4,7 +4,7 @@
  * This file is part of Libgcrypt.
  *
  * Libgcrypt is free software; you can redistribute it and/or modify
- * it under the terms of the GNU Lesser general Public License as
+ * it under the terms of the GNU Lesser General Public License as
  * published by the Free Software Foundation; either version 2.1 of
  * the License, or (at your option) any later version.
  *
@@ -24,9 +24,6 @@
 #define G10_CIPHER_PROTO_H
 
 
-enum pk_encoding;
-
-
 /* Definition of a function used to report selftest failures.
    DOMAIN is a string describing the function block:
           "cipher", "digest", "pubkey or "random",
@@ -184,7 +181,7 @@ typedef struct gcry_cipher_spec
   } flags;
   const char *name;
   const char **aliases;
-  gcry_cipher_oid_spec_t *oids;
+  const gcry_cipher_oid_spec_t *oids;
   size_t blocksize;
   size_t keylen;
   size_t contextsize;
@@ -219,15 +216,13 @@ typedef void (*gcry_md_final_t) (void *c);
 typedef unsigned char *(*gcry_md_read_t) (void *c);
 
 /* Type for the md_extract function.  */
-typedef void (*gcry_md_extract_t) (void *c, void *outbuf, size_t nbytes);
-
-/* Type for the md_hash_buffer function. */
-typedef void (*gcry_md_hash_buffer_t) (void *outbuf, const void *buffer,
-				       size_t length);
+typedef gpg_err_code_t (*gcry_md_extract_t) (void *c, void *outbuf,
+					     size_t nbytes);
 
 /* Type for the md_hash_buffers function. */
-typedef void (*gcry_md_hash_buffers_t) (void *outbuf, const gcry_buffer_t *iov,
-				        int iovcnt);
+typedef void (*gcry_md_hash_buffers_t) (void *outbuf, size_t nbytes,
+					const gcry_buffer_t *iov,
+					int iovcnt);
 
 typedef struct gcry_md_oid_spec
 {
@@ -243,16 +238,15 @@ typedef struct gcry_md_spec
     unsigned int fips:1;
   } flags;
   const char *name;
-  unsigned char *asnoid;
+  const unsigned char *asnoid;
   int asnlen;
-  gcry_md_oid_spec_t *oids;
+  const gcry_md_oid_spec_t *oids;
   int mdlen;
   gcry_md_init_t init;
   gcry_md_write_t write;
   gcry_md_final_t final;
   gcry_md_read_t read;
   gcry_md_extract_t extract;
-  gcry_md_hash_buffer_t hash_buffer;
   gcry_md_hash_buffers_t hash_buffers;
   size_t contextsize; /* allocate this amount of context */
   selftest_func_t selftest;
